# Copyright 2014, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#    Josef Skladanka <jskladan@redhat.com>

import os
import sys
from optparse import OptionParser

from alembic.config import Config
from alembic import command as al_command
from alembic.migration import MigrationContext

from vault import db
from vault.models.user import User
from vault.models.secrets import Secret, Bucket, BucketRoles


def get_alembic_config():
    # the location of the alembic ini file and alembic scripts changes when
    # installed via package
    if os.path.exists("./alembic.ini"):
        alembic_cfg = Config("./alembic.ini")
    else:
        alembic_cfg = Config("/usr/share/vault/alembic.ini",
                             ini_section='alembic-packaged')
    return alembic_cfg


def upgrade_db(*args):
    alembic_cfg = get_alembic_config()

    context = MigrationContext.configure(db.engine.connect())
    current_rev = context.get_current_revision()
    print "Upgrading Database to `head` from `%s`" % current_rev

    al_command.upgrade(alembic_cfg, "head")


def init_alembic(*args):
    alembic_cfg = get_alembic_config()

    # check to see if the db has already been initialized by checking for an
    # alembic revision
    context = MigrationContext.configure(db.engine.connect())
    current_rev = context.get_current_revision()

    if not current_rev:
        print "Initializing alembic"
        print " - Setting the version to the first revision"
        al_command.stamp(alembic_cfg, "15f5eeb9f635")
    else:
        print "Alembic already initialized"


def initialize_db(destructive):
    alembic_cfg = get_alembic_config()

    print "Initializing database"
    if destructive:
        print " - Dropping all tables"
        db.drop_all()
        print " - Creating tables"
        db.create_all()
        print " - Stamping alembic's current version to 'head'"
        al_command.stamp(alembic_cfg, "head")

    init_alembic()
    upgrade_db()


def mock_data(destructive):
    print "Populating tables with mock-data"

    if destructive:
        print " - Users"
        joe = User('joe', 'joe')
        jane = User('jane', 'jane')

        db.session.add(joe)
        db.session.add(jane)

        print " - Secrets"

        bucket = Bucket("Joe's bucket", joe)
        bucket.uuid = 'uuid_1'
        secret = Secret(bucket, "surname", "Doe")
        role1 = BucketRoles(bucket, jane, 'client')
        db.session.add(secret)

        bucket = Bucket("Jane's bucket", jane)
        bucket.uuid = 'uuid_2'
        secret = Secret(bucket, "password", "jane_doe_123", 1024)
        db.session.add(secret)

        db.session.commit()
    else:
        print " - skipped Secrets"


def test_data(*args):
    for s in db.session.query(Secret).all():
        print "%s | %s | %s" % (s.bucket.name, s.name, s.value)


def main():
    possible_commands = ['init_db', 'mock_data', 'upgrade_db', 'init_alembic', 'test_data']

    usage = 'usage: [DEV=true] %prog ' + "(%s)" % ' | '.join(possible_commands)
    parser = OptionParser(usage=usage)
    parser.add_option("-d", "--destructive",
                      action="store_true", dest="destructive", default=False,
                      help="Drop tables in `init_db`; Store data in `mock_data` "
                      "even if the tables are not empty")

    (options, args) = parser.parse_args()

    if len(args) != 1 or args[0] not in possible_commands:
        print usage
        print
        print 'Please use one of the following commands: %s' % str(possible_commands)
        sys.exit(1)

    command = {
        'init_db': initialize_db,
        'mock_data': mock_data,
        'upgrade_db': upgrade_db,
        'init_alembic': init_alembic,
        'test_data': test_data,
    }[args[0]]

    if not options.destructive:
        print "Proceeding with non-destructive init. To perform destructive "\
            "steps use -d option."

    command(options.destructive)

if __name__ == '__main__':
    main()
