# Copyright 2014, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#    Josef Skladanka <jskladan@redhat.com>

import flask
from flask import Blueprint, render_template, redirect, flash, url_for, request
from flask.ext.wtf import Form
from wtforms import TextField, PasswordField, HiddenField
from wtforms.validators import Required
import flask_login
from flask.ext.login import login_user, logout_user
from flask.ext.login import current_user, AnonymousUserMixin

import munch

from vault import app, login_manager, db, OIDC
from vault.models.user import User
from vault.models.user import USERNAME_OIDC_GROUP_PREFIX

login_page = Blueprint('login_page', __name__)


class LoginForm(Form):
    username = TextField(u'Username', validators=[Required()])
    password = PasswordField(u'Password', validators=[Required()])
    next_page = HiddenField()


def get_oidc_data():
    flask.session.permanent = True
    if OIDC.user_loggedin:
        # The "sub" value is unique per user, and does not make an HTTP API request.
        user_sub = OIDC.user_getfield("sub")
        if ('oidc' not in flask.session or not flask.session['oidc'] or flask.session['oidc'].get("sub") != user_sub):
            user_info = OIDC.user_getinfo(["name", "nickname", "sub", "email", "zoneinfo", "cla", "groups",])
            flask.session['oidc'] = dict(
                    sub=user_info['sub'],
                    email=user_info['email'],
                    groups=user_info['groups'],
                    )
            # Changes on mutable objects aren't picked up:
            # http://flask.pocoo.org/docs/0.12/api/#flask.session.modified
            flask.session.modified = True
    else:
        flask.session['oidc'] = {}

def load_user_privileges(user):
    get_oidc_data()
    if not flask.session.get('oidc'):
        return user

    user.bucket_group_roles = []
    # FIXME: do this in a saner way in the future. Cache the existing groups, probably
    existing_groups = User.query.filter(User.username.like('%s%%' % USERNAME_OIDC_GROUP_PREFIX)).all()
    existing_groups = {grp.username:grp for grp in existing_groups}
    for user_group in flask.session['oidc']['groups']:
        groupname = "%s%s" % (USERNAME_OIDC_GROUP_PREFIX, user_group)
        if groupname in existing_groups:
            group = existing_groups[groupname]
            user.bucket_group_roles.extend(group.bucket_roles)
        else:
            group = User(groupname, None, 'oidc_group')
            db.session.add(group)

        db.session.commit()

    return user

@login_manager.user_loader
def load_user(userid):
    user = User.query.get(userid)
    print "USER:", user
    if user:
        user = load_user_privileges(user)
    else:
        user = AnonymousUserMixin
    return user


@login_page.route("/oidc_login")
@OIDC.require_login
def oidc_login():
    get_oidc_data()
    email = flask.session.get('oidc', {}).get('email')
    if not email:
        return redirect(url_for('login_page.logout'))

    user = User.query.filter_by(username=email).first()
    if not user:
        user = User(email, None)
        db.session.add(user)
        db.session.commit()
    # FIXME: do this in a saner way in the future. Cache the existing groups, probably
    existing_groups = User.query.filter(User.username.like('%s%%' % USERNAME_OIDC_GROUP_PREFIX)).all()
    existing_groups = [grp.username for grp in existing_groups]
    for user_group in flask.session['oidc']['groups']:
        groupname = "%s%s" % (USERNAME_OIDC_GROUP_PREFIX, user_group)
        if groupname not in existing_groups:
            group = User(groupname, None, 'oidc_group')
            db.session.add(group)
        db.session.commit()

    login_user(load_user_privileges(user))

    return redirect(url_for('main.show_buckets'))


@login_page.route('/login', methods=['GET', 'POST'])
def login():
    login_form = LoginForm()

    if login_form.validate_on_submit():
        user = User.query.filter_by(username=login_form.username.data).first()
        if user and user.check_password(login_form.password.data):
            login_user(user)

            app.logger.info('Successful login for user %s' % login_form.username.data)
            flash('Logged In Successfully!')

            return redirect(login_form.next_page.data)
        else:
            app.logger.info('FAILED login for user %s' % login_form.username.data)
            flash('Login Failed! Please Try again!')

    login_form.next_page.data = request.args.get('next') or url_for('main.show_buckets')
    return render_template('login.html', form=login_form)


@login_page.route('/logout')
def logout():
    OIDC.logout()
    logout_user()
    flask.session.clear()
    flash('Logged Out Successfully!')
    return redirect(url_for('main.index'))
