# Copyright 2014, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#    Josef Skladanka <jskladan@redhat.com>

import re

from flask import Blueprint, render_template, request, jsonify, redirect, flash, url_for, abort, request
import werkzeug.exceptions
from sqlalchemy.orm import exc as orm_exc

from flask.ext.restful import reqparse
from flask.ext.login import login_user, logout_user, login_required, current_user
from werkzeug.exceptions import HTTPException
from werkzeug.exceptions import BadRequest as JSONBadRequest

from vault import app, db
from vault.models.user import User
from vault.models.secrets import Secret, Bucket, BucketRoles
from vault.serializers.api_v1 import Serializer

main = Blueprint('main', __name__)

# TODO: find out why error handler works for 404 but not for 400
@app.errorhandler(400)
def bad_request(error):
    return jsonify({"message": "Bad request"}), 400


@app.errorhandler(404)
def not_found(error):
    return jsonify({"message": "Not found"}), 404


# =============================================================================
#                               GLOBAL VARIABLES
# =============================================================================

QUERY_LIMIT = 20

RE_PAGE = re.compile(r"([?&])page=([0-9]+)")
RE_CALLBACK = re.compile(r"([?&])callback=[^&]*&?")
RE_CLEAN_AMPERSANDS = re.compile(r'&+')

# RP contains request parsers (reqparse.RequestParser).
#    Parsers are added in each 'resource section' for better readability
RP = {}

SERIALIZE = Serializer().serialize


# =============================================================================
#                               GLOBAL METHODS
# =============================================================================

def pagination(q, page, limit):
    """
    Sets the offset/limit for the DB query.
    limit+1 is purposely set as 'limit' so we can later on decide whether 'next'
    page link should be provided or set to None.
    """
    # pagination offset
    try:
        page = int(page)
        if page > 0:
            offset = page * limit
            q = q.offset(offset)
    except (TypeError, ValueError):
        pass

    q = q.limit(limit + 1)
    return q


def prev_next_urls(data, limit=QUERY_LIMIT):
    global RE_PAGE

    try:
        match = RE_PAGE.findall(request.url)
        flag, page = match[0][0], int(match[0][1])
    except IndexError:  # page not found
        page = None

    prev = None
    next = None
    placeholder = "[!@#$%^&*PLACEHOLDER*&^%$#@!]"

    if page is None:
        if "?" in request.url:
            baseurl = "%s&page=%s" % (request.url, placeholder)
        else:
            baseurl = "%s?page=%s" % (request.url, placeholder)
        page = 0
    else:
        baseurl = RE_PAGE.sub("%spage=%s" % (flag, placeholder), request.url)

    baseurl = RE_CALLBACK.sub(r"\1", baseurl)
    baseurl = RE_CLEAN_AMPERSANDS.sub('&', baseurl)

    if page > 0:
        prev = baseurl.replace(placeholder, str(page - 1))
    if len(data) > limit:
        next = baseurl.replace(placeholder, str(page + 1))
        data = data[:limit]

    return data, prev, next

# =============================================================================
#                               UI
# =============================================================================

@main.route('/')
@main.route('/index')
def index():
    return render_template("index.html")


@main.route('/buckets', methods=['GET'])
@login_required
def show_buckets():
    owner = [r.bucket for r in current_user.bucket_roles if r.bucket.active and r.role == 'owner']
    admin = [r.bucket for r in current_user.bucket_roles if r.bucket.active and r.role == 'admin']
    client = [r.bucket for r in current_user.bucket_roles if r.bucket.active and r.role == 'client']

    if current_user.bucket_group_roles:
        owner.extend([r.bucket for r in current_user.bucket_group_roles if r.bucket.active and r.role == 'owner'])
        admin.extend([r.bucket for r in current_user.bucket_group_roles if r.bucket.active and r.role == 'admin'])
        client.extend([r.bucket for r in current_user.bucket_group_roles if r.bucket.active and r.role == 'client'])


    return render_template("show_buckets.html", owner=owner, admin=admin, client=client)



from flask.ext.wtf import Form
from wtforms import Label, TextField, SelectField, TextAreaField, HiddenField, FieldList, FormField
from wtforms.validators import Required


class RoleEntryForm(Form):
    username = TextField(u'Username')
    role = SelectField(u'Role', choices=[
            ('client', "Client"),
            ('admin',  "Admin"),
            ('owner', "Owner"),
            ('remove', "Remove"),
        ])

class SecretEntryForm(Form):
    secret_id = HiddenField(default="-1")
    sname = TextField(u'name')
    svalue = TextField(u'value')


class BucketForm(Form):
    bucket_id = HiddenField()
    bucket_name = TextField(u'bucket name')
    bucket_desc = TextField(u'bucket description')
    current_roles = FieldList(FormField(RoleEntryForm))
    new_roles = FieldList(FormField(RoleEntryForm), min_entries=2)
    current_secrets = FieldList(FormField(SecretEntryForm))
    new_secrets = FieldList(FormField(SecretEntryForm), min_entries=4)


@main.route('/buckets/<bucket_uuid>/edit', methods=['GET', 'POST'])
@login_required
def edit_bucket(bucket_uuid):
    bucket = db.session.query(Bucket).filter_by(uuid=bucket_uuid).first()

    if not bucket:
        abort(404)

    editor_has_admin = False
    editor_has_owner = False
    if current_user in bucket.admins:
        editor_has_admin = True
    if current_user in bucket.owners:
        editor_has_owner = True
    if not (editor_has_admin or editor_has_owner):
        abort(403)

    if request.method == 'GET':
        owners = [r for r in bucket.roles if r.role == 'owner']
        admins = [r for r in bucket.roles if r.role == 'admin']
        clients = [r for r in bucket.roles if r.role == 'client']

        role_data = []
        for r in owners:
            role_data.append({'username': r.user.username, 'role': 'owner'})
        for r in admins:
            role_data.append({'username': r.user.username, 'role': 'admin'})
        for r in clients:
            role_data.append({'username': r.user.username, 'role': 'client'})

        secret_data = []
        for s in bucket.secrets:
            if not s.active:
                continue
            secret_data.append({'sname': s.name, 'svalue': s.ciphertext})

        form = BucketForm(current_roles=role_data, current_secrets=secret_data)
        form.bucket_id.data = bucket.id
        form.bucket_name.data = bucket.name
        form.bucket_desc.data = bucket.description

        return render_template("edit_bucket.html",
                form=form,
                form_action=url_for('main.edit_bucket', bucket_uuid=bucket_uuid),
                bucket_name=bucket.name)

    if request.method == 'POST':
        validation_ok = True
        form = BucketForm()


        # BUCKETS
        bucket_has_owner = False
        seen_users = []
        # FIXME - do this better... for some reason FieldList can not do []+[]
        form_roles = [f for f in form.current_roles]
        form_roles.extend([f for f in form.new_roles])
        for f in form_roles:
            if not f.username.data.strip():
                continue
            f.username.data = f.username.data.strip()

            user = User.query.filter_by(username=f.username.data).first()
            if not user:
                # FIXME: check FAS whether user/group exists instead of failing
                flash("User '%s' not found" % f.username.data, 'danger')
                validation_ok = False
                continue
            if user in seen_users:
                flash("Only one entry per user allowed for a bucket. Found '%s' multiple times." % user.username, 'danger')
                validation_ok = False
                continue

            seen_users.append(user)
            try:
                user_role = [x for x in bucket.roles if x.user_id == user.id].pop()
            except IndexError:
                user_role = BucketRoles(bucket, user, None)

            # Skip "non-changes"
            if user_role.role == f.role.data:
                if user_role.role == 'owner':
                    bucket_has_owner = True
                continue

            if f.role.data == 'remove':
                if user_role.role == 'owner' and not editor_has_owner:
                    flash("You need to have owner access to remove owners", 'danger')
                    continue
                if user_role.role is None: # this happens when 'new user' is added, but role is set to 'remove'. Do nothing.
                    continue
                print user_role
                bucket.roles.remove(user_role)
                db.session.delete(user_role)
                continue

            if f.role.data == 'owner':
                if not editor_has_owner:
                    flash("You need to have owner access to set owner privileges for %s" % user.username, 'danger')
                    validation_ok = False
                    continue

                bucket_has_owner = True
                user_role.role = 'owner'
            else: # the selected role can only be admin or client, just set it.
                user_role.role = f.role.data

            db.session.add(user_role)

        if not bucket_has_owner:
            validation_ok = False
            flash("No owner set for bucket", 'danger')


        # SECRETS
        # FIXME - do this better... for some reason FieldList can not do []+[]
        form_secrets = [f for f in form.current_secrets]
        form_secrets.extend([f for f in form.new_secrets])
        seen_secrets = []
        for f in form_secrets:
            if not f.sname.data.strip():
                continue
            f.sname.data = f.sname.data.strip()

            if f.sname.data in seen_secrets:
                flash("Only one entry per secret is allowed. Found '%s' multiple times" % f.sname.data, 'danger')
                validation_ok = False
                continue
            seen_secrets.append(f.sname.data)

            if not f.svalue.data:
                flash("Missing value for secret '%s'" % s.sname.data, 'danger')
                validation_ok = False
                continue

            try:
                secret = [x for x in bucket.secrets if x.active and x.name == f.sname.data].pop()
            except IndexError:
                db.session.add(Secret(bucket, f.sname.data, f.svalue.data))
                continue

            if f.svalue.data.strip() == secret.ciphertext:
                continue

            secret.active = False
            db.session.add(Secret(bucket, secret.name, f.svalue.data, secret.version + 1))
            db.session.add(secret)


        if validation_ok:
            bucket.name = form.bucket_name.data
            bucket.description = form.bucket_desc.data
            db.session.add(bucket)
            db.session.commit()
            return redirect(url_for('main.show_buckets'))
        else:
            db.session.rollback()
            return render_template("edit_bucket.html",
                    form=form,
                    form_action=url_for('main.edit_bucket', bucket_uuid=bucket_uuid),
                    bucket_name=bucket.name)

@main.route('/buckets/add')
@login_required
def add_bucket():
    bucket = Bucket("New bucket", current_user)
    db.session.add(bucket)
    db.session.commit()
    return redirect(url_for('main.edit_bucket', bucket_uuid = bucket.uuid))

# =============================================================================
#           API - FIXME - probably use OIDC api tokens?
# =============================================================================

from flask_httpauth import HTTPBasicAuth
http_auth = HTTPBasicAuth()

@http_auth.verify_password
def verify_pw(username, password):
    user = User.query.filter_by(username=username).first()
    if user and user.check_password(password):
        login_user(user)
        return True


@main.route('/api/v1/buckets', methods=["GET"])
@http_auth.login_required
def list_buckets():
    data = [r.bucket for r in current_user.bucket_roles if r.bucket.active]
    return jsonify(dict(
        data=[SERIALIZE(o) for o in data],
    ))


@main.route('/api/v1/buckets/<bucket_uuid>', methods=["GET"])
@http_auth.login_required
def get_bucket(bucket_uuid):
    bucket = db.session.query(Bucket).filter_by(uuid=bucket_uuid).first()
    if not bucket:
        return jsonify({'message': 'Bucket not found'}), 404
    if not current_user in [r.user for r in bucket.roles]: # FIXME make this better!
        return jsonify({'message': 'Forbidden'}), 403
    return jsonify(SERIALIZE(bucket))


@main.route('/api/v1/buckets/<bucket_uuid>/secrets', methods=["GET"])
@http_auth.login_required
def get_bucket_secrets(bucket_uuid):
    bucket = db.session.query(Bucket).filter_by(uuid=bucket_uuid).first()
    if not bucket:
        return jsonify({'message': 'Bucket not found'}), 404
    if not current_user in [r.user for r in bucket.roles]: # FIXME make this better!
        return jsonify({'message': 'Forbidden'}), 403
    data = dict([(s.name, s.value) for s in bucket.secrets if s.active])
    return jsonify(data)

