# Copyright 2014, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#    Josef Skladanka <jskladan@redhat.com>

import os
import hashlib
import base64
import uuid as lib_uuid


from Crypto.Cipher import AES
from Crypto import Random

from vault import db, app
from vault.serializers import DBSerialize
from .user import User

pad = lambda s: s + (AES.block_size - len(s) % AES.block_size) * chr(AES.block_size - len(s) % AES.block_size)
unpad = lambda s: s[:-ord(s[len(s)-1:])]

MASTERKEY = base64.b64decode(app.config['MASTERKEY'])

def encrypt(value, key):
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    ciphertext = (iv + cipher.encrypt(pad(value)))
    return base64.b64encode(ciphertext)


def decrypt(ciphertext, key):
    ciphertext = base64.b64decode(ciphertext)
    dec_iv = ciphertext[:AES.block_size]
    dec_cipher = AES.new(key, AES.MODE_CBC, dec_iv)
    return unpad(dec_cipher.decrypt(ciphertext[16:]))


class Secret(db.Model, DBSerialize):
    id = db.Column(db.Integer, primary_key=True)
    bucket_id = db.Column(db.Integer, db.ForeignKey('bucket.id'))

    name = db.Column(db.String(80))
    ciphertext = db.Column(db.Text)
    version = db.Column(db.Integer)
    active = db.Column(db.Boolean, default=True)

    bucket = db.relation('Bucket', backref='secrets')

    def __init__(self, bucket, name, value, version=1):
        self.name = name
        self.bucket = bucket
        self.ciphertext = encrypt(value, self.bucket.access_key)
        self.version = version

    @property
    def value(self):
        if not hasattr(self, '_value'):
            self._value = decrypt(self.ciphertext, self.bucket.access_key)
        return self._value


    @value.setter
    def value(self, new_value):
        self._value = new_value
        self.ciphertext = encrypt(new_value, self.bucket.access_key)

class Bucket(db.Model, DBSerialize):
    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String(36), unique=True)
    name = db.Column(db.String(80))
    ciphertext = db.Column(db.Text)
    active = db.Column(db.Boolean, default=True)
    description = db.Column(db.Text)

    __table_args__ = (
        db.Index('bucket_idx_uuid', 'uuid',
                 postgresql_ops={'uuid': 'text_pattern_ops'},
                 ),
    )

    def __init__(self, name, owner, description=None):
        self.uuid = str(lib_uuid.uuid1())
        self.name = name
        self.description = description
        self.ciphertext = encrypt(hashlib.pbkdf2_hmac('sha256', os.urandom(256), os.urandom(16), 100000), MASTERKEY)
        self.roles.append(BucketRoles(self, owner, 'owner'))

    @property
    def access_key(self):
        if not hasattr(self, '_access_key'):
            self._access_key = decrypt(self.ciphertext, MASTERKEY)
        return self._access_key

    @property
    def owners(self):
        return [r.user for r in self.roles if r.role=='owner']

    @property
    def admins(self):
        return [r.user for r in self.roles if r.role=='admin']

    @property
    def clients(self):
        return [r.user for r in self.roles if r.role=='client']


class BucketRoles(db.Model, DBSerialize):
    id = db.Column(db.Integer, primary_key=True)
    bucket_id = db.Column(db.Integer, db.ForeignKey('bucket.id'))
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'))

    role = db.Column(db.String(10), default=None)

    bucket = db.relation('Bucket', backref='roles')
    user = db.relation('User', backref='bucket_roles')

    def __init__(self, bucket, user, role):
        self.bucket = bucket
        self.user = user
        self.role = role

    def __str__(self):
        return "<BucketRoles: %s, %s, %s>" % (self.bucket.name, self.user.username, self.role)
